/*
*      This program is free software: you can redistribute it and/or modify
*      it under the terms of the GNU General Public License as published by
*      the Free Software Foundation, either version 2 of the License, or
*      (at your option) any later version.
*
*      Copyright (C) 2020 iopsys Software Solutions AB
*		Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
*/

#include <string.h>
#include <sys/stat.h>
#include <ctype.h>
#include <uci.h>
#include "udpechoserveruci.h"
#include "udpechoserverlog.h"

static struct uci_context *uci_ctx = NULL;
static struct uci_context *uci_ctx_state = NULL;

int dmuci_init(void)
{
	uci_ctx = uci_alloc_context();
	if (!uci_ctx) {
		return -1;
	}
	uci_ctx_state = uci_alloc_context();
	if (!uci_ctx_state) {
		return -1;
	}
	uci_add_delta_path(uci_ctx_state, uci_ctx_state->savedir);
	uci_set_confdir(uci_ctx_state, VAR_STATE);
	uci_set_savedir(uci_ctx_state, NULL);
	return 0;
}

int dmuci_fini(void)
{
	if (uci_ctx) {
		uci_free_context(uci_ctx);
	}
	if (uci_ctx_state) {
		uci_free_context(uci_ctx_state);
	}
	uci_ctx = NULL;
	uci_ctx_state = NULL;
	return 0;
}

static bool dmuci_validate_section(const char *str)
{
	if (!*str)
		return false;

	for (; *str; str++) {
		unsigned char c = *str;

		if (isalnum(c) || c == '_')
			continue;

		return false;
	}
	return true;
}

static int dmuci_init_ptr(struct uci_context *ctx, struct uci_ptr *ptr, char *package, char *section, char *option, char *value)
{
	memset(ptr, 0, sizeof(struct uci_ptr));

	/* value */
	if (value) {
		ptr->value = value;
	}
	ptr->package = package;
	if (!ptr->package)
		goto error;

	ptr->section = section;
	if (!ptr->section) {
		ptr->target = UCI_TYPE_PACKAGE;
		goto lastval;
	}

	ptr->option = option;
	if (!ptr->option) {
		ptr->target = UCI_TYPE_SECTION;
		goto lastval;
	} else {
		ptr->target = UCI_TYPE_OPTION;
	}

lastval:
	if (ptr->section && !dmuci_validate_section(ptr->section))
		ptr->flags |= UCI_LOOKUP_EXTENDED;

	return 0;

error:
	return -1;
}

void dmuci_print_list(struct uci_list *uh, char **val, char *delimiter)
{
	struct uci_element *e;
	static char buffer[512];
	char *buf = buffer;
	*buf = '\0';

	uci_foreach_element(uh, e) {
		if (*buf) {
			strcat(buf, delimiter);
			strcat(buf, e->name);
		}
		else {
			strcpy(buf, e->name);
		}
	}
	*val = buf;
}

char *dmuci_get_value(char *package, char *section, char *option)
{
	struct uci_ptr ptr;
	char *val = "";

	if (!section || !option)
		return val;

	if (dmuci_init_ptr(uci_ctx, &ptr, package, section, option, NULL)) {
		return val;
	}
	if (uci_lookup_ptr(uci_ctx, &ptr, NULL, true) != UCI_OK) {
		return val;
	}

	if (!ptr.o)
		return val;

	if(ptr.o->type == UCI_TYPE_LIST) {
		dmuci_print_list(&ptr.o->v.list, &val, " ");
		return val;
	}

	if (ptr.o->v.string)
		return ptr.o->v.string;
	else
		return val;
}

char *dmuci_set_value(char *package, char *section, char *option, char *value)
{
	struct uci_ptr ptr;
	int ret = UCI_OK;

	if (!section)
		return "";

	if (dmuci_init_ptr(uci_ctx, &ptr, package, section, option, value)) {
		return "";
	}
	if (uci_lookup_ptr(uci_ctx, &ptr, NULL, true) != UCI_OK) {
		return "";
	}

	ret = uci_set(uci_ctx, &ptr);

	if (ret == UCI_OK) {
		ret = uci_save(uci_ctx, ptr.p);
		uci_commit(uci_ctx, &ptr.p, false);
	}

	if (ptr.o && ptr.o->v.string)
		return ptr.o->v.string;

	return "";
}

/************************* /var/state ***************************/
char *dmuci_set_value_state(char *package, char *section, char *option, char *value)
{
	char *val;
	struct uci_context *save_uci_ctx = uci_ctx;
	uci_ctx = uci_ctx_state;
	val = dmuci_set_value(package, section, option, value);
	uci_ctx = save_uci_ctx;
	return val;
}

char *dmuci_get_value_state(char *package, char *section, char *option)
{
	char *val;
	struct uci_context *save_uci_ctx = uci_ctx;
	uci_ctx = uci_ctx_state;
	val = dmuci_get_value(package, section, option);
	uci_ctx = save_uci_ctx;
	return val;
}

bool varstate_section_add(const char *section)
{
	struct uci_ptr ptr;
	struct uci_section *s = NULL;
	memset(&ptr, 0, sizeof(struct uci_ptr));

	ptr.package = "udpechoserver";
	ptr.target = UCI_TYPE_PACKAGE;

	dmuci_init();
	if (uci_lookup_ptr(uci_ctx_state, &ptr, NULL, false) != UCI_OK) {
		udpechoserver_log(SINFO,"%s lookup ptr !OK", section);
		dmuci_fini();
		return true;
	}

	s = uci_lookup_section(uci_ctx_state, ptr.p, section);
	if (s != NULL) {
		udpechoserver_log(SINFO,"%s section already present", section);
		dmuci_fini();
		return true;
	}

	if (uci_add_section(uci_ctx_state, ptr.p, section, &s) != UCI_OK || !s) {
		dmuci_fini();
		udpechoserver_log(SINFO,"failed to add %s section", section);
		return false;
	}

	ptr.package = s->package->e.name;
	ptr.p = s->package;
	ptr.section = s->e.name;
	ptr.s = s;
	ptr.value = section;
	ptr.target = UCI_TYPE_SECTION;

	if (uci_lookup_ptr(uci_ctx_state, &ptr, NULL, true) == UCI_OK) {
		uci_rename(uci_ctx_state, &ptr);
		if (uci_commit(uci_ctx_state, &ptr.p, false) != UCI_OK) {
			udpechoserver_log(SINFO,"failed to commit");
			dmuci_fini();
			return false;
		}
	}

	udpechoserver_log(SINFO,"added %s section", section);
	dmuci_fini();
	return true;
}

void reset_udpecho_stat(void)
{
	struct uci_section *s = NULL;
	struct uci_ptr ptr;
	memset(&ptr, 0, sizeof(struct uci_ptr));

	ptr.package = "udpechoserver";
	ptr.target = UCI_TYPE_PACKAGE;

	dmuci_init();
	if (uci_lookup_ptr(uci_ctx_state, &ptr, NULL, false) != UCI_OK) {
		udpechoserver_log(SINFO,"package not found");
		dmuci_fini();
		return;
	}

	s = uci_lookup_section(uci_ctx_state, ptr.p, "statistics");
	if (s == NULL) {
		udpechoserver_log(SINFO,"statistics section not present");
		dmuci_fini();
		return;
	}

	ptr.package = s->package->e.name;
	ptr.section = s->e.name;
	ptr.option = NULL;
	ptr.value = NULL;
	ptr.flags |= UCI_LOOKUP_COMPLETE;

	ptr.p = s->package;
	ptr.s = s;
	ptr.last = &ptr.s->e;
	ptr.target = UCI_TYPE_SECTION;

	if (uci_delete(uci_ctx_state, &ptr) != UCI_OK) {
		udpechoserver_log(SINFO,"failed to delete statistics section");
		dmuci_fini();
		return;
	}

	uci_commit(uci_ctx_state, &ptr.p, false);

	udpechoserver_log(SINFO,"statistics section deleted");
	dmuci_fini();
	return;
}

void config_var_state_uci(void)
{
	char *value = NULL;
	bool enabled = false;
	struct stat buf;

	// existance check
	char filename[] = "/var/state/udpechoserver";
	int exist = stat(filename, &buf);
	udpechoserver_log(SINFO,"fileexist %d", exist);
	if (exist != 0) {
		udpechoserver_log(SINFO,"create file %s", filename);
		FILE *fp = fopen(filename, "w");
		if (!fp) {
			udpechoserver_log(SINFO,"Failed to create file %s", filename);
			return;
		}
		fclose(fp);
	}

	udpechoserver_log(SINFO,"Reset stat");
	reset_udpecho_stat();

	if (!varstate_section_add("udpechoserver")) {
		return;
	}

	dmuci_init();
	value = dmuci_get_value("udpechoserver", "udpechoserver", "plus");
	if (value != NULL && *value != '\0') {
		if ((strcasecmp(value,"true")==0) || (strcmp(value,"1")==0)) {
			enabled = true;
		}
	}

	dmuci_set_value_state("udpechoserver", "udpechoserver", "plus", enabled ? "1" : "0");
	dmuci_fini();

	if (!varstate_section_add("statistics")) {
		return;
	}

	return;
}
